 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN getMSDPtr,USBErrorFeatureDisabled,Pop2USBErrorNotEnoughMem,DriverBuf,DoAssert

;Organization of cache
;bytes        function
;1            max # of cache entries
;1            cache flags
;1            base LRU
;2            pointer to LRU counters
;2            pointer to sector buffers
;size*4       sector #s of cache entries
;size*2       lru counters
;size*512     sector buffers


CacheSize            EQU 9D95h                   ;1 a
CacheFlags           EQU CacheSize+1             ;1 a
CacheBaseLRU         EQU CacheFlags+1            ;1
CacheLRUPtr          EQU CacheBaseLRU+1          ;2
CacheDataPtr         EQU CacheLRUPtr+2           ;2
CacheHits            EQU CacheDataPtr+2          ;2
CacheMisses          EQU CacheHits+2             ;2
CacheSectorNumbers   EQU CacheMisses+2           ;2

CacheOverheadBytes EQU CacheSectorNumbers-CacheSize

cacheFlagRAM EQU 0


 IF CACHE_ON = 1
SaveToCache:
; Prepare to add data to cache
;Input
; MSDLBA = block to read
;Output 
; DE -> buffer to load data to
       LOG    Cache,logCacheSaveToCache

       call   IsCacheOn
       ret    z
       
       call   DumpBlock
       ld     b,a
       call   ResetLRU

;       push   de
       push   bc


       ld     l,b
       ld     h,0
       add    hl,hl
       add    hl,hl
       ld     bc,CacheSectorNumbers
       add    hl,bc                       ;hl-> sector # for block B
              

       call   GetMSDPtrDE          ;DE->MSDLBA
       ex     de,hl
       ldi
       ldi
       ldi
       ldi                         ;Saved the sector #



       pop    bc
       call   GetCacheDataOffset
;       pop    de
;       ld     bc,512
       ex     de,hl
;       ldir

       ret

GetMSDPtrDE:
	push hl
	ld hl,(MemPtr)
	ld de,offMSDMemPtr
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
       inc    de                   ;->MSDLBA
       pop    hl
       ret       

DumpBlock:
;Find block with the highest LRU
;Return block # in A
;Preserve BC, DE, HL
       push   hl
       push   de
       push   bc

       ld     a,(CacheSize)
       ld     b,a
       ld     hl,(CacheLRUPtr)

       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       inc    hl
       ld     c,b
       dec    b
       jr     z,DumpDone

FindDumpLoop:
       push   bc
       ld     c,(hl)
       inc    hl
       ld     b,(hl)                      ;bc = current lru
       inc    hl
       ex     de,hl
       or     a
       push   hl
       sbc    hl,bc
       pop    hl
       ex     de,hl
       jr     nc,NotNewDump
       ld     d,b
       ld     e,c
       pop    bc
       ld     c,b
       push   bc
NotNewDump:
       pop    bc
       djnz   FindDumpLoop
DumpDone:

       ld     a,(CacheSize)
       sub    c
       pop    bc
       pop    de
       pop    hl
       ret



ResetLRU:
;Set LRU for block B to 0 and increment other LRUs
;Preserve BC, DE, HL
       LOG    Cache,logCacheResetLRU
       push   bc
       push   de
       push   hl



       ld     a,(CacheSize)        ;a = max cache size

       ld     hl,(CacheLRUPtr)

                                   ;B = block # to reset
       ld     c,a                  ;C = max # of blocks
       xor    a                    ;A = current block #
ResetLRULoop:
       ld     e,(hl)
       inc    hl
       ld     d,(hl)               ;DE = current LRU
       inc    de                   ;increment
       cp     b                    ; unless it's the one we're resetting
       jr     nz,$f
;       ld     de,0                 ; in which case set to 0
       ld     de,(CacheBaseLRU)
       ld     d,0
$$:
       ld     (hl),d
       dec    hl
       ld     (hl),e
       inc    hl
       inc    hl
       inc    a
       cp     c
       jr     nz,ResetLRULoop

       pop    hl
       pop    de
       pop    bc

       ret


GetCacheDataOffset:
;Get data ptr for a block
;Input
; B = block #
;Output
; HL = ptr to block
;Preserves BC, DE
       push   bc
       push   de

       ld     l,b
       ld     h,0


       add    hl,hl
       add    hl,hl
       add    hl,hl

       add    hl,hl
       add    hl,hl
       add    hl,hl
       
       add    hl,hl
       add    hl,hl
       add    hl,hl         ;HL * 512

       ld     de,(CacheDataPtr)
       
       add    hl,de

       pop    de
       pop    bc
       ret       

ReadFromCache:
;Read a block of data from cache
;Input
; B = block #
; DE -> where to read to (MUST BE <0C000h)
;Should return NC
       LOG    Cache,logCacheReadFromCache 
       call   IsCacheOn
       ret    z
       call   ResetLRU
       call   GetCacheDataOffset
;       ld     bc,512
;       ldir
       ex     de,hl
       or     a
       ret





UpdateCache:
;Update cache contents on write
;Must preserve DE and MSDLBA
;Input
; DE = buffer to write
; MSDLBA = sector #

       call   IsCacheOn
       ret    z
       call   IsInCache
       ret    c
;       call   InvalidateEntry
       push   de
       call   ResetLRU
       call   GetCacheDataOffset
       ld     bc,512
       ex     de,hl
       ldir
       pop    de

       ret

 COMMENT ~
InvalidateEntry:
;Invalidate a cache entry (sets sector # to 80... and LRU to 8080)
;Input
; B = entry # to invalidate
;Preserves DE, HL,
       LOG    Custom,7
       LOG    Custom,b
       push   hl
       push   de
       ld     a,80h                ;Invalidation #
       ld     l,b
       ld     h,0
       push   hl
       add    hl,hl
       ld     de,(CacheLRUPtr)
       add    hl,de         ;hl -> LRU for this block
       ld     (hl),a
       inc    hl
       ld     (hl),a
       


       pop    hl                   ;HL = block #
       add    hl,hl
       add    hl,hl
       ld     de,CacheSectorNumbers
       add    hl,de                ;HL-> sector # for this block
       

       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a

       pop    de
       pop    hl
       ret
 ~

IsInCache:
;Is MSDLBA in cache
;Input
; MSDLBA = sector to look for
;Output
; NC = sector is in cache
; B = sector # in cache
; Preserves DE, HL, IX
       in     a,(7)
       ASSERTNZ 81h,a
       LOG    Cache,logCacheIsInCache
       call   IsCacheOn
       ret    z
       push   hl
       push   de
       push   ix
       call   getMSDPtr                   ;ix -> msd data mem
       ld     a,(CacheSize)
       ld     b,a                         ;B = size of cache
       

       push   ix
       pop    de
       inc    de                          ;DE -> MSDLBA

       ld     hl,CacheSectorNumbers
       push   bc
IsInCacheLoop:       
       push   bc
       push   de
       push   hl
       ld     b,4
$$:
       ld     a,(de)
       cp     (hl)
       jr     nz,NotThisSector
       inc    de
       inc    hl
       djnz   $b
       pop    hl
       pop    de
       pop    bc                          ; b = loop #
       pop    af                          ; a = cache size
       sub    b
       ld     b,a                         ;b = cache buffer #
       pop    ix
       pop    de
 IF LOGGING_ON = 1
       ld     hl,(CacheHits)
       inc    hl
       ld     (CacheHits),hl
 ENDIF
       pop    hl
       or     a
       ret

NotThisSector:
       pop    hl
       ld     de,4
       add    hl,de
       pop    de
       pop    bc
       djnz   IsInCacheLoop
       pop    bc
       pop    ix
       pop    de
 IF LOGGING_ON = 1
       ld     hl,(CacheMisses)
       inc    hl
       ld     (CacheMisses),hl
 ENDIF
       pop    hl
       scf
       ret



       




IsCacheOn:
;Check if cache is active
;Preserve all
;NZ = active
       push   hl
       push   de
       ld     hl,(MemPtr)
       ld     de,offFlags
       add    hl,de
       bit    flagCache,(hl)
       pop    de
       pop    hl
       ret


CalcCacheBytes:
;Input
; C = number of sectors to cache
;Output
; HL = size of buffer to allocate
       push   bc
       ld     l,c
       ld     h,0
       push   hl
       add    hl,hl
       add    hl,hl
       ld     h,130
       B_CALL HTimesL
       pop    de
       add    hl,de
       add    hl,de
       ld     de,CacheOverheadBytes
       add    hl,de
       pop    bc
       ret

 ENDIF

CacheOn:
;Start caching
;Input
; B = flags
;  bit 0 - reset = use InsertMem to allocate RAM in user memory
;            set = user ram page in D with port 28h to swap in memory
; C = cache size (# of sectors)
; D = RAM page to use if flags.0 = set
; E = base LRU value for non FAT pages
;Output
; HL = size of cache memory allocated


 IF CACHE_ON = 0
       jp     USBErrorFeatureDisabled      
 ELSE
       LOG    Cache,logCacheOn
       call   IsCacheOn
       ret    nz
  IF LOGGING_ON = 1
       ld     hl,0
       ld     (CacheHits),hl
       ld     (CacheMisses),hl
  ENDIF






       bit    cacheFlagRAM,b
       jr     z,$f
       ld     a,76h
       out    (28h),a
       ld     a,d
       out    (7),a
       jr     CacheOnCont
$$:
       push   de
       push   bc
       call   CalcCacheBytes       

       push   hl
       B_CALL EnoughMem
       pop    hl
       jp     c,Pop2USBErrorNotEnoughMem
       ld     de,9D95h
       B_CALL InsertMem
       pop    bc
       pop    de
CacheOnCont:

       ld     (CacheSize),bc
       ld     a,e
       ld     (CacheBaseLRU),a

       ld     l,c
       ld     h,0
       add    hl,hl
       add    hl,hl
       ld     de,CacheOverheadBytes
       add    hl,de
       ld     de,9D95h
       add    hl,de
       ld     (CacheLRUPtr),hl
       ld     e,a
       ld     d,0
       add    hl,de
       add    hl,de
       ld     (CacheDataPtr),hl


       ld     b,a
       ld     hl,CacheSectorNumbers
       ld     a,080h

$$:
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       ld     (hl),a
       inc    hl
       djnz   $b                   ;Initialize sector numbers and LRU counts
       
       ld     hl,(MemPtr)
       ld     de,offFlags
       add    hl,de
       set    flagCache,(hl)
       ld     a,(CacheSize)
       ld     c,a
       call   CalcCacheBytes
       ret
 ENDIF
      
       
CacheOff:
;turn off cache

;       ld     a,85h
;       out    (5),a
;       ld     hl,9D95h
;       ld     de,0C000h
;       ld     bc,3000h
;       ldir
;       xor    a
;       out    (5),a

 IF CACHE_ON = 0
       ret
 ELSE

       LOG    Cache,logCacheOff 
       call   IsCacheOn
       ret    z
       ld     a,(CacheFlags)
       bit    cacheFlagRAM,a
       jr     z,$f
       ld     a,81h
       out    (7),a
       xor    a
       out    (28h),a
       jr     CacheOffCont              
$$:
       ld     a,(CacheSize)
       ld     c,a
       call   CalcCacheBytes
       ex     de,hl
       ld     hl,9D95h
       B_CALL DelMem
CacheOffCont:

       ld     hl,(MemPtr)
       ld     de,offFlags
       add    hl,de
       res    flagCache,(hl)
 IF LOGGING_ON = 1
  IF LOGCACHE_ON = 1
       LOG    DataStart,4
       ld     hl,(CacheHits)
       LOG    Data,h
       LOG    Data,l
       ld     hl,(CacheMisses)
       LOG    Data,h
       LOG    Data,l
       LOG    DataEnd
  ENDIF
 ENDIF
       ret
 ENDIF


