 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN USBErrorNoDevice,IsDeviceConnected,USBErrorInit,USBErrorClass,CDBuf,USBReqData
 EXTERN USBErrorAutoSetup,GetClass,AutoSetup,NegHL,USBHostInit,DispHexA,ReadDescriptor,PopRetC,RetC
 EXTERN FindDescriptor,PopErrorReadDescriptor,PopErrorFindDescriptor, AutoSetupNoInit,IPutC
 EXTERN GetIDs,USBErrorID,SendData,SetCallBack,SendControlOut,DoAssert,SendControlOutSafe,AutoSetupSafe
 EXTERN GetClassSafe,KillUSBDevice,USBErrorFeatureDisabled

 

KBDVersion:
;Output
; DE = size of buffer required for keyboard driver
; H = major version
; L = minor version
       ld     de,kbdMemSize
       ld     hl,KVersion
       ret


KBDInit:
 IF KBD_ON = 0
       jp     USBErrorFeatureDisabled  
 ELSE
       call   KBDInitDo
       ret    nc
       call   KillUSBDevice
       scf
       ret

KBDInitDo:
       DISAVE
       call   KBDInitSafe
       EIRESTORE
       ret

KBDInitSafe:
;Initialize KBD driver
;Input
; HL -> kbd memory buffer
; DE -> temp buffer for descriptors
; IX -> callback routine
; B = KBD flags
;  0 - reset = auto handle lock keys
;      set   = don't handle lock keys (pass back to user)
;  1 -       = Initial state of caps lock
;  2 -       = Initial state of num lock
;  3 -       = Initial state of scroll lock
;  4 - set   = Pass through callback
;Output
; NC = success

       LOG    KBDInit,b
       push   ix
       ld     ix,(MemPtr)
       ld     (ix+offKBDBuf),l
       ld     (ix+offKBDBuf+1),h     
       ld     a,b
       ld     bc,offkbdFlags
       add    hl,bc
       ld     (hl),a               ;Store KBDFlags
       dec    hl
       ld     (hl),0               ;Set Q size to 0
       ld     bc,offkbdCallBack-offkbdQSize
       add    hl,bc
       pop    bc
       ld     (hl),c
       inc    hl
       ld     (hl),b               ;Set callback passthrough
       ld     hl,KBDCallBack
       ex     de,hl
       push   hl                   ;Save desc buf loc
       call   AutoSetupSafe
       pop    hl
       ret    c
       call   GetClassSafe
       ret    c
       cp     3
       jr     nz,USBErrorClass
       dec    b
       jr     nz,USBErrorClass
       dec    c
       jr     nz,USBErrorClass
       call   KBDSendLocks
       ld     b,1
       call   USBReqData
       or     a
       ret
 ENDIF


KBDCheckDirect:
;Check if key A is currently being pressed
;Returns Z if the key is being pressed
 IF KBD_ON = 0
       ret
 ELSE
       DISAVE
       ld     hl,(MemPtr)
       inc    hl
       inc    hl
       ld     b,6
$$:
       cp     (hl)
       jr     z,CheckDirectDone
       inc    hl
       djnz   $b
       xor    a
       inc    a
CheckDirectDone:
       EIRESTORE
       ret
 ENDIF

KBDGetMods:
;Gets current state of modifiers
;Output
; A = shift keys
; B = lock keys
 IF KBD_ON = 0
       ret
 ELSE
       call   LdDEDataBuf
       ld     hl,offkbdFlags
       add    hl,de
       ld     a,(hl)
       ld     hl,(MemPtr)
       ld     b,(hl)
       ret
 ENDIF

KBDGetKey:
;Retrieves a key from the keypress queue
;Output
; NC = success
; A = # of elements current remaining in kbd queue
; B = raw key modifiers
; C = raw key press
; D = converted TIOS key extend token
; E = converted TIOS key token
 IF KBD_ON = 0
       ret
 ELSE
       DISAVE
       LOG    KBDGetKey
       call   LdDEDataBuf
       ld     hl,offkbdQ
       add    hl,de                ;hl-> kbdQ
       ld     d,h
       ld     e,l                  ;de-> kbdQ
       ld     bc,offkbdQSize-offkbdQ
       add    hl,bc                ;hl-> kbdQSize
       push   hl
       pop    ix                   ;ix+1 -> kbdFlags
       ld     a,(hl)
       ASSERTNC 6,a
       or     a
       jr     z,EIRetC               ;Nothing in Q
       dec    (hl)
       ld     a,(hl)               ;A = remaining Q size
       ex     de,hl                ;hl-> kbdQ
       push   hl
       ld     b,(hl)
       inc    hl
       ld     c,(hl)
       inc    hl                   ;hl-> kbdQ+2
       pop    de                   ;de-> kbdQ
       push   bc
       ld     bc,kbdMaxQ*2-2
       ldir                        ;update Q
       pop    bc    
       push   bc


       call   KBDTranslate
       push   bc
       pop    de
       pop    bc

       EIRESTORE
       or     a
       LOG    Custom,a
       LOG    Custom,b
       LOG    Custom,c
       LOG    Custom,d
       LOG    Custom,e
       ret

EIRetC:
       EIRESTORE
       scf
       ret


DoNumLock:
;Adjust keypress in C based on num log flag in (ix+1)
       bit    kbdFlagNumLock,(ix+1)
       ret    nz
       ld     a,c
       res    7,a
       sub    kbdKN1
       cp     11
       ret    nc
       push   hl
       ld     hl,KeyTabNumLock
       ld     e,a
       ld     d,0
       add    hl,de
       ld     c,(hl)
       pop    hl
       ret

DoCapsLock:
;Adjust keypress in C based on caps lock flag in (ix+1)

       bit    kbdFlagCapsLock,(ix+1)
       ret    z
       call   IsLetter
       ret    nc
       set    7,c
       ld     hl,KeyTabUpper
       ret
 ENDIF

KBDTranslate:
;Translate from raw keys to TIOS keypresses
;Input 
; B = modifier keys
; C = key value
; IX+1 -> kbdFlags
;Output
; B = key extend value
; C = key code
;Must preserve: A
 IF KBD_ON = 0
       ret
 ELSE
       push   af
       ld     hl,KeyTabLower
       ld     a,b                  ;modifiers
       ld     b,0                  ;default is no keyextend
       and    kbdLShift+kbdRShift
       jr     z,$f
       set    7,c                  ;Indicate shift
       ld     hl,KeyTabUpper
$$:

       call   DoCapsLock
       call   DoNumLock

       call   CheckExtend          ;Set keyextend in B as necessary

       ld     e,c
       res    7,e                  ;Remove shift bit
       ld     d,0
       add    hl,de
       ld     c,(hl)
       pop    af
       ret

CheckExtend:
       push   hl
       push   bc
       ld     e,0                  ;current extend value

       call   IsLetter
       jr     nc,NotLowerLet
       bit    7,c
       jr     nz,NotLowerLet
       ld     e,0FCh               ;It's a lower case letter so set key extend
NotLowerLet:
       ld     hl,KeyTabFB
       ld     b,KeyTabFBSize
       ld     d,0FBh
       call   CheckInTab
;       ld     hl,KeyTabFC       No current FC keys
;       ld     b,KeyTabFCSize
;       ld     d,0FCh
;       call   CheckInTab
       ld     hl,KeyTabFE
       ld     b,KeyTabFESize
       ld     d,0FEh
       call   CheckInTab
       pop    bc
       ld     b,e
       pop    hl
       ret
CheckInTab:
       ld     a,(hl)
       cp     c
       jr     z,$f
       inc    hl
       djnz   CheckInTab
       ret
$$:
       ld     e,d
       ret

IsLetter:
;Returns carry if raw key value in C is a letter
       push   bc
       ld     b,a
       ld     a,c
       res    7,a
       sub    kbdKA
       cp     26
       ld     a,b
       pop    bc
       ret



KBDCallBack:
       ld     b,6                  ;# of keys reported by kbd
       ld     hl,(MemPtr)          ;HL-> modifier keys
       ASSERTC h,9Dh
       ASSERTC 7Fh,h
       inc    hl                   ;HL-> reserved
       inc    hl                   ;HL-> keypresses
       call   KBDEmptyBuf          ;remove keys that are no longer being pressed
KBDCBLoop:
       ld     a,(hl)
       call   KBDInBuf             ;Check if key was already being pressed
       jr     z,$f                 ; If so, don't report it as a new keypress
       call   KBDAddBuf            ; If not, add keypress to buffer
       call   KBDAddToQ            ; And pass it on to key queue
$$:
       inc    hl
       djnz   KBDCBLoop

       call   LdDEDataBuf
       ld     hl,offkbdFlags
       add    hl,de
       bit    kbdFlagCallBack,(hl)
       jr     z,$f
       ld     de,offkbdCallBack-offkbdFlags
       add    hl,de
       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       push   de
       ASSERTC d,9Dh
       ASSERTC 7Fh,d
$$:
       ld     b,1
       jp     USBReqData           ;Get new data



KBDAddToQ:
;Add keypress at (hl) to kbd Q

       push   hl
       call   KBDAddToQLocks
       jr     z,KBDAddToQDone       

       call   LdDEDataBuf
       ld     hl,offkbdQ
       add    hl,de                ;hl-> kbdQ
       push   hl
       ld     de,offkbdQSize-offkbdQ
       add    hl,de                ;hl-> kbdQSize
       ld     a,(hl)               ;A = kbdQSize
       ASSERTNC 6,a
       ASSERTC h,9Dh
       ASSERTC 7Fh,h
       pop    de                   ;de-> kbdQ
       cp     kbdMaxQ
       jr     z,KBDAddToQDone      ;Q is full
       inc    (hl)                 ;increment Q size

       add    a,a
       ld     l,a
       ld     h,0

       add    hl,de                ;HL-> place to add in Q

       ld     de,(MemPtr)
       ASSERTC d,9Dh
       ASSERTC 7Fh,d
       ld     a,(de)               ;modifiers
       ld     (hl),a
       inc    hl
       pop    de
       push   de
       ld     a,(de)               ;cur keypress
       ld     (hl),a
       
KBDAddToQDone:       
       pop    hl

       ret


KBDAddToQLocks:
;Handle lock keys
;Z if abort add to Q
       ld     a,(hl)               ;current key press
       call   LdDEDataBuf
       ld     hl,offkbdFlags
       add    hl,de
       bit    kbdFlagNoLocks,(hl)
       ret    nz
       ld     d,1<<kbdFlagCapsLock
       cp     kbdKCapsLock
       jr     z,$f
       ld     d,1<<kbdFlagNumLock
       cp     kbdKNumLock
       jr     z,$f
       ld     d,1<<kbdFlagScrollLock
       cp     kbdKScrollLock
       jr     z,$f
       ret
$$:
       ld     a,d
       xor    (hl)
       ld     (hl),a
       call   KBDSendLocks
       xor    a
       ret





KBDEmptyBuf:
       push   hl
       push   bc
;       ld     de,KBDBuf
       call   LdDEDataBuf
       ld     b,6
KBDEmptyBufLoop:
       ld     a,(de)
       ld     c,6
;       ld     hl,DriverBuf+2
       ld     hl,(MemPtr)
       inc    hl
       inc    hl
$$:
       or     a
       jr     z,$f
       cp     (hl)
       jr     z,$f
       inc    hl
       dec    c
       jr     nz,$b
       xor    a
       ld     (de),a
$$:
       inc    de
       djnz   KBDEmptyBufLoop
       pop    bc
       pop    hl
       ret

KBDInBuf:
       or     a
       ret    z
;       ld     de,KBDBuf
       call   LdDEDataBuf
       ld     c,6
$$:
       ld     a,(de)
       cp     (hl)
       ld     a,(hl)
       ret    z
       inc    de
       dec    c
       jr     nz,$b
       inc    c
       ret
KBDAddBuf:
;       ld     de,KBDBuf
       call   LdDEDataBuf
       ld     c,6
KBDAddBufLoop:
       ld     a,(de)
       or     a
       jr     nz,$f
       ld     a,(hl)
       ld     (de),a
       ret
$$:
       inc    de
       dec    c
       jr     nz,KBDAddBufLoop
       ret


LdDEDataBuf:
       push   hl
       ld     hl,(MemPtr)
       ASSERTC h,9Dh
       ASSERTC 7Fh,h
       ld     de,offKBDBuf
       add    hl,de
       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       ASSERTC d,9Dh
       ASSERTC 7Fh,d
       pop    hl
       ret

KBDSendLocks:
;Send locks status to keyboard
       push   bc
       push   hl
       push   de
       B_CALL ZeroOP1
       call   LdDEDataBuf
       ld     hl,offkbdFlags
       add    hl,de
       ld     a,(hl)
       rra
       and    7
       ld     (OP1),a
       ld     hl,KBDSetReport
       ld     de,OP1
;       ld     b,8
       ld     b,1
       call   SendControlOutSafe
       pop    de
       pop    hl
       pop    bc
       ret


KBDSetReport:
       db     21h,9,00,02,00,00,1,00

       db     3


KeyTabNumLock:
       db     kbdKEnd
       db     kbdKDown
       db     kbdKPageDown
       db     kbdKLeft
       db     0
       db     kbdKRight
       db     kbdKHome
       db     kbdKUp
       db     kbdKPageUp
       db     kbdKInsert
       db     kbdKDelete

KeyTabFB:
       db     kbdK2+128
       db     kbdK3+128
       db     kbdK4+128
       db     kbdK5+128
       db     kbdK7+128
       db     kbdKSemicolon
       db     kbdKBackSlash
       db     kbdKBackSlash+128
       db     kbdKMinus+128
       db     kbdKBackQuote
       db     kbdKBackQuote+128
KeyTabFBSize EQU $-KeyTabFB

KeyTabFE:
       db     kbdK1+128
       db     kbdKEqu
       db     kbdKPeriod+128
       db     kbdKComma+128
       db     kbdKApostrophe
KeyTabFESize EQU $-KeyTabFE

;KeyTabFC:
;KeyTabFCSize EQU $-KeyTabFC


KeyTabLower:
       db     0,0,0,0
       db     kLa,kLb,kLc,kLd,kLe,kLf,kLg,kLh,kLi,kLj,kLk,kLl,kLm,kLsmalln,kLo,kLp,kLq,kLsmallr,kLs,kLt,kLu,kLv
       db     kLw,kLx,kLy,kLz
       db     k1,k2,k3,k4,k5,k6,k7,k8,k9,k0
       db     kEnter,kQuit,21h,kSpace
       db     kSpace,kSub,kTEqu,kLBrack,kRBrack,kBackSlash
       db     0
       db     kSemicolon,kApost,kBackQuote,kComma,kDecPnt,kDiv
       db     0
       db     kYEqu,kWindow,kZoom,kTrace,kGraph,0,0,0,0,0,0,0
       db     0,0,0,kIns,kBOL
       db     kAlphaUp,kDel,kEOL,kAlphaDown
       db     kRight,kLeft,kDown,kUp,0
       db     kDiv,kMul,kSub,kAdd
       db     kEnter
       db     k1,k2,k3,k4,k5,k6,k7,k8,k9,k0,kDecPnt
       db     0,0,0,0,0,0,0
       db     0,0,0,0,0,0,0
       db     0,0,0,0,0,0,0


KeyTabUpper:
       db     0,0,0,0
       db     kCapA,kCapB,kCapC,kCapD,kCapE,kCapF,kCapG,kCapH,kCapI,kCapJ,kCapK,kCapL,kCapM
       db     kCapN,kCapO,kCapP,kCapQ,kCapR,kCapS,kCapT,kCapU,kCapV,kCapW,kCapX,kCapY,kCapZ
       db     kXFactorial,kAtSign,kPound,kDollar,kPercent,kExpon,kAmpersand,kMul,kLParen,kRParen
       db     kEnter,kQuit,21h,kSpace
       db     kSpace,kUnderscore,kAdd,kLBrace,kRBrace,kVertSlash
       db     0
       db     kColon,kQuote,kTilde,kTLT,kTGT,kQuest
       db     0
       db     kStatP,kTblSet,kFormat,kCalc,kTable,0,0,0,0,0,0,0
       db     0,0,0,kIns,kBOL
       db     kAlphaUp,kDel,kEOL,kAlphaDown
       db     kRight,kLeft,kDown,kUp,0
       db     kDiv,kMul,kSub,kAdd
       db     kEnter
       db     k1,k2,k3,k4,k5,k6,k7,k8,k9,k0,kDecPnt
       db     0,0,0,0,0,0,0
       db     0,0,0,0,0,0,0
       db     0,0,0,0,0,0,0

 ENDIF
