;USB Flash interface routines

 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN getMSDPtr,sendCmdIn,sendCmdOut,DispHexA,SetupLog,StopLog,getMSDData,DispHexHL,getSectorBuffer,IGetKey,USBErrorCont,Disp4HL
 EXTERN UpdateCache,IsInCache,ReadFromCache,SaveToCache,IsCacheOn,CacheBaseLRU,sendReset,WaitTimerB,USBReqData,ILdHLInd,SendControlData

UFI_Initialize:
	;Initializes USB device for sector read/write operations
	;Input:      None
	;Output:     Initializes USB flash drive
	;Returns C if error (including no FATxx partition)
	;Only checks 4 primary partitions, selects the first FAT12/16/32 partition found
 IF (FAT12_ON = 1) || (FAT32_ON = 1)
       ld hl,(MemPtr)
       ld de,offFlags
       add hl,de
       res flagIsFAT12,(hl)
       res flagIsFAT32,(hl)
 ENDIF
       call UFI_getCapacity
       ret c
	call getMSDPtr
	;ld l,(ix+MSDBPS) ;value is now hard-coded
	;ld h,(ix+MSDBPS+1)
	;ld de,512
	;or a
	;sbc hl,de ;checks for sectors larger than 512 bytes
	;ld a,errMSDLargeSect
	;jr c,USBErrorCont
	ld (ix+MSDLBA),0
	ld (ix+MSDLBA+1),0
	ld (ix+MSDLBA+2),0
	ld (ix+MSDLBA+3),0
	call getSectorBuffer
	call UFI_Read
	ret c
       ld de,(cachetemp)
       ld hl,36h
       add hl,de
       ld de,UFI_FAT16
       push hl
       call cmpstr5
       pop hl
       jr z,UFI_Continue
 IF FAT12_ON = 1
       ld de,UFI_FAT12
       call cmpstr5
       jr z,UFI_isFAT12
 ENDIF
 IF FAT32_ON = 1
       ld de,(cachetemp)
       ld hl,52h
       add hl,de
       ld de,UFI_FAT32
       call cmpstr5
       jr z,UFI_isFAT32
 ENDIF
       ld de,(cachetemp)
       ld b,4
	ld hl,1C2h
	add hl,de
UFI_PartitionTable:
	ld a,(hl)
       cp 0Eh
       jr z,UFI_PartitionTable_OK ;FAT16
	cp 04h
	jr z,UFI_PartitionTable_OK ;FAT16 (<32 MB)
	cp 06h
	jr z,UFI_PartitionTable_OK ;FAT16 (32 MB - 2 GB)
 IF FAT32_ON = 1
	cp 0Bh
	jr z,UFI_PartitionTable_FAT32
	cp 0Ch
	jr z,UFI_PartitionTable_FAT32
 ENDIF
 IF FAT12_ON = 1
       dec a
       jr z,UFI_PartitionTable_FAT12
 ENDIF
$$:	ld de,10h
	add hl,de
	djnz UFI_PartitionTable
	ld a,errMSDNoParts
	jr USBErrorCont
 IF FAT12_ON = 1
UFI_PartitionTable_FAT12: ;preserve HL
       push hl
       call getFlagAddress
       set flagIsFAT12,(hl)
       pop hl
       jr UFI_PartitionTable_OK
 ENDIF
 IF FAT32_ON = 1
UFI_PartitionTable_FAT32: ;preserve HL
       push hl
       call getFlagAddress
       set flagIsFAT32,(hl)
       pop hl
 ENDIF
UFI_PartitionTable_OK:
	;in LBA mode, relative sectors is the starting LBA block
	ld bc,4
	add hl,bc
	;Relative sectors
$$:    push hl
	ld hl,MSDPartStart
	call getMSDData
	ex de,hl
	pop hl
	ldir
	ret
IsFAT12:
 IF FAT12_ON = 1
       push de
       push hl
       call getFlagAddress
       bit flagIsFAT12,(hl)
       pop hl
       pop de
       ret
UFI_isFAT12:
       call getFlagAddress
       set flagIsFAT12,(hl)
       jr UFI_Continue
UFI_FAT12:
       DB "FAT12"
 ENDIF
IsFAT32:
 IF FAT32_ON = 1
       push de
       push hl
       call getFlagAddress
       bit flagIsFAT32,(hl)
       pop hl
       pop de
       ret
UFI_FAT32:
       DB "FAT32"
UFI_isFAT32:
       call getFlagAddress
       set flagIsFAT32,(hl)
 ENDIF
UFI_Continue:
       call getSectorBuffer
       ld hl,1Ch
       add hl,de
       ld bc,4
       jr $B
getFlagAddress:
       ld hl,(MemPtr)
       ld de,offFlags
       add hl,de
       ret
cmpstr5:
       ld b,5
cmpstrb:
$$:    ld a,(de)
       cp (hl)
       ret nz
       inc hl
       inc de
       djnz $B
       xor a
       ret
UFI_FAT16:
       DB "FAT16"

UFI_Write:
	;Writes one sector of data from sector buffer
	;Input:	(MSDLBA) contains the starting address
	;		DE is the location to read the data to be written from
	;Output:	Writes one block
	;Returns C if error
 IF CACHE_ON = 1
       call UpdateCache
 ENDIF
	call getMSDPtr
	ld (ix+MSDDataInAddr),e ;really DataOutAddr
	ld (ix+MSDDataInAddr+1),d
	ld e,(ix+MSDBPS)
	ld d,(ix+MSDBPS+1)
	ld (ix+MSDDataInExpected),e
	ld (ix+MSDDataInExpected+1),d
	ld hl,MSDPacketData
	call getMSDData
	push hl
	ld (hl),0Ah ;previously 9
	inc hl
	ld (hl),2Ah
	inc hl
	ld (hl),0
	inc hl
	ld a,(ix+MSDLBA+3)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA+2)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA+1)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA)
	ld (hl),a
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	ld (hl),1
       inc hl
       ld (hl),0
	pop hl
       ei ;this shouldn't be necessary
	jp sendCmdOut

UFI_ReadPriority:
 IF CACHE_ON = 1
       call   IsCacheOn
       jr     z,SkipCache
       ld     a,(CacheBaseLRU)
       push   af
       xor    a
       ld     (CacheBaseLRU),a
       call   UFI_Read
       pop    bc
       ld     a,b
       ld     (CacheBaseLRU),a
       ret
 ENDIF

; fall through to UFI_Read, do not reorder

UFI_Read:
	;Reads one sector of data into sector buffer
	;Input:	(MSDLBA) contains the starting address
	;             DE is the location to store the block of data
	;Output:	Reads one block
	;Returns C if error
       ;If caching is on, returns DE -> location data was read to (actually, it always does this)
 IF CACHE_ON = 1
       call   IsCacheOn
       jr     z,SkipCache

       call   IsInCache
       jr     nc,UFI_ReadFromCache
       call   SaveToCache
SkipCache:
 ENDIF      
       ld (cachetemp),de
	call getMSDPtr
	ld (ix+MSDDataInAddr),e
	ld (ix+MSDDataInAddr+1),d
       push de
	ld e,(ix+MSDBPS)
	ld d,(ix+MSDBPS+1)
	ld (ix+MSDDataInExpected),e
	ld (ix+MSDDataInExpected+1),d
	ld hl,MSDPacketData
	call getMSDData
	push hl
	ld (hl),0Ah ;previously 9
	inc hl
	ld (hl),28h
	inc hl
	ld (hl),0
	inc hl
	ld a,(ix+MSDLBA+3)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA+2)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA+1)
	ld (hl),a
	inc hl
	ld a,(ix+MSDLBA)
	ld (hl),a
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	ld (hl),1
       inc hl
       ld (hl),0
	pop hl
       ei ;this shouldn't be necessary
	call sendCmdIn
       pop de
       ret
 IF CACHE_ON = 1
UFI_ReadFromCache:
       call   ReadFromCache
       ld     (cachetemp),de
       ret
 ENDIF

UFI_getCapacity:
	;Sets up (MSDBPS)
	;Input:      None
	;Output:     (MSDBPS) contains bytes-per-sector
	;C returned if error
	;Destroys OP2
       COMMENT ~ This is the routine to mess with when analyzing a particular drive
       ld hl,MSDPacketData
       call getMSDData
       call getMSDPtr
       ld (ix+MSDDataInAddr),l
       ld (ix+MSDDataInAddr+1),h
       ld hl,UFI_UnitReady
       ld de,0
       ld (ix+MSDDataInExpected),e
       ld (ix+MSDDataInExpected+1),d
       ei
       call sendCmdIn
       ld a,0F0h
       jr c,USBErrorCont
       ld hl,MSDPacketData
       call getMSDData
       call getMSDPtr
       ld (ix+MSDDataInAddr),l
       ld (ix+MSDDataInAddr+1),h
       ld hl,UFI_Capacity
       ld de,0FCh
       ld (ix+MSDDataInExpected),e
       ld (ix+MSDDataInExpected+1),d
       ei
       call sendCmdIn
       ld a,0FEh
       jr c,USBErrorCont
       ~
       ld hl,MSDPacketData ;send "request sense" since some drives freak without it
       call getMSDData
       call getMSDPtr
       ld de,512
       ld (ix+MSDBPS),e ;hard-coded BPS
       ld (ix+MSDBPS+1),d
       ld (ix+MSDDataInAddr),l
       ld (ix+MSDDataInAddr+1),h
       ld hl,UFI_RequestSense
       ld de,12h
       ld (ix+MSDDataInExpected),e
       ld (ix+MSDDataInExpected+1),d
       ei
       jr sendCmdIn ;call sendCmdIn
       ;ld a,0FFh ;TEST CODE
       ;jr c,USBErrorCont
       ;ret
UFI_RequestSense:
       DB 0Ch,03h,0,0,0,12h,0,0,0,0,0,0
       COMMENT ~ TEST STRINGS
UFI_LUN:
       DB 0A1h,0FEh,0,0,0,0,1,0
UFI_Capacity:
       DB 0Ch,23h,0,0,0,0,0,0,0FCh,0,0,0
UFI_UnitReady:
       DB 0Ch,00h,0,0,0,0,0,0,0,0,0,0
       ~

       COMMENT ~ GetCapacity packet not accepted by some drives
	call getMSDPtr
	ld hl,MSDPacketData
	call getMSDData
	ld (ix+MSDDataInAddr),l
	ld (ix+MSDDataInAddr+1),h
	ld de,8
	ld (ix+MSDDataInExpected),e
	ld (ix+MSDDataInExpected+1),d
	ld hl,UFI_getCapacity_UFI
       ei ;shouldn't be, and maybe not, necessary
	call sendCmdIn
       ret c
	ld hl,MSDPacketData
	call getMSDData
	push hl
	pop ix
	ld h,(ix)
	ld l,(ix+1)
	push ix
	B_CALL SetXXXXOP2
	ld hl,UFI_getCapacity_64k
	rst 20h
	B_CALL FPMult
	pop ix
	ld h,(ix+2)
	ld l,(ix+3)
	B_CALL SetXXXXOP2
	B_CALL FPAdd
	ld hl,MSDPacketData
	call getMSDData
	ld de,6
	add hl,de
	push hl
	pop ix
	ld h,(ix)
	ld l,(ix+1)
	call getMSDPtr
	ld (ix+MSDBPS),l
	ld (ix+MSDBPS+1),h
	B_CALL SetXXXXOP2
	B_CALL FPMult
	or a
	ret
UFI_getCapacity_UFI:
	DB 0Ah,25h,0,0,0,0,0
UFI_getCapacity_64k:
	DB 00,84h,65h,53h,60h,0,0,0,0
       ~

	COMMENT ~ This function isn't used for any reason right now
UFI_getName:
	;Gets vendor/product ID strings from USB flash drive
	;Input:	None
	;Output:	DE points to a vendor identification string, null-terminated
	;		HL points to a product identification string, null-terminated
       ;Notes:       This routine might make assumptions it shouldn't
	;Destroys OP1/OP2/OP3
	;C returned if error
	call getMSDPtr
	ld hl,MSDPacketData
	call getMSDData
	ld (ix+MSDDataInAddr),l
	ld (ix+MSDDataInAddr+1),h
	ld (ix+MSDDataInExpected),36
	ld hl,UFI_getName_UFI
	call sendCmdIn
	ret c
	ld hl,MSDPacketData
	call getMSDData
	ld de,8
	add hl,de
	push hl
	rst 20h
	xor a
	ld (OP1+8),a
	pop hl
	ld de,8
	add hl,de
	ld de,OP2
	ld bc,16
	ldir
	ld (de),a
	ld hl,OP2
	ld de,OP1
	or a
	ret
UFI_getName_UFI:
	DB 6,12h,0,0,0,24h,0
	~
