;Code to deal with hubs

 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN SendControlData,DriverInit,AutoSetup,USBReqData,DispHexA,IPutC,KillUSBDevice,KillDriver,WaitTimer16
 EXTERN ReadDescriptor,SetupLog,PumpOn,GetClass,RetC,SetAddressSafe,GetMaxPacketSizeSafe

 IF HUBCODE_ON=1
SetHubPortFeature:
       LOG    SetPortFeature,a
       LOG    SetPortFeature,b
       ld     hl,SetPortFeatureReqData
       jr     $f
ClearHubPortFeature:
;Send request to clear hub port feature
; A = port #
; B = feature type
       LOG    ClearPortFeature,a
       LOG    ClearPortFeature,b
       ld     hl,ClearPortFeatureReqData
$$:       
       push   bc
       ld     de,OutBuf
       ld     bc,8
       push   de
       ldir
       pop    hl
       ld     (OutBuf+4),a
       pop    af
       ld     (OutBuf+2),a
       ld     b,0
       call   SendControlData
       ret


SetHubFeature:
       LOG    SetHubFeature,a
       ld     hl,SetHubFeatureReqData
       jr     $f
ClearHubFeature:
;Send request to clear hub feature
; A = feature type
       LOG    ClearHubFeature,a
       ld     hl,ClearHubFeatureReqData
$$:
       ld     de,OutBuf
       ld     bc,8
       push   de
       ldir
       pop    hl

       ld     (OutBuf+2),a
       ld     b,0
       call   SendControlData
       ret    c
;       xor    a                    ;*  These are from setaddress.  Not sure if needed
;       out    (8Eh),a              ;*
;       out    (91h),a              ;*
       ret

GetHubStatus:
;Load Hub Status into DE (4 byte buffer)
; NC = success
       LOG    GetHubStatus
       ld     hl,GetHubStatusReqData
       ld     b,4
       jp     SendControlData

GetHubPortStatus:
;Load Port Status for port A into DE (4 byte buffer)
; NC = success
       LOG    GetHubPortStatus,a
       push   de
       ld     de,OutBuf
       ld     hl,GetPortStatusReqData
       ld     bc,8
       push   de
       ldir
       pop    hl
       ld     (OutBuf+4),a
       ld     b,4
       pop    de
       jp     SendControlData



ClearHubFeatureReqData:
       db     20h,1,0,0,0,0,0,0
ClearPortFeatureReqData:
       db     23h,1,0,0,0,0,0,0
SetHubFeatureReqData:
       db     20h,3,0,0,0,0,0,0
SetPortFeatureReqData:
       db     23h,3,0,0,0,0,0,0

GetHubStatusReqData:
       db     0A0h,0,0,0,0,0,4,0
GetPortStatusReqData:
       db     0A3h,0,0,0,0,0,4,0


ConfigureHub:
       push   hl
       call   GetClass
       pop    hl
       ret    c
       cp     9
       jr     nz,RetC
       push   hl
       ld     b,descHub
       call   ReadDescriptor
       pop    hl
       ret    c
       inc    hl
       inc    hl
       ld     b,(hl)               ;Number of hub ports
FindConnectedPortLoop:
       push   bc
       ld     a,b
       ld     b,hubfeatPortPower
       call   SetHubPortFeature
       pop    bc
       ret    c
       call   WaitTimer16
       push   bc
       ld     a,b
       ld     b,hubfeatPortReset
       call   SetHubPortFeature
       pop    bc
       ret    c
       call   WaitTimer16
       push   bc
       ld     de,OP1
       ld     a,b
       call   GetHubPortStatus
       pop    bc
       ret    c
       ld     a,(OP1)
       bit    0,a
       jr     nz,FoundConnectedPort
       push   bc
       ld     a,b
       ld     b,hubfeatPortEnable
       call   ClearHubPortFeature
       pop    bc
       djnz   FindConnectedPortLoop
       scf
       ret

FoundConnectedPort:


       xor    a
       out    (80h),a
       out    (5Bh),a
       call   GetMaxPacketSizeSafe
       jp     SetAddressSafe




HubTest:
       B_CALL ClrLCDFull
       B_CALL HomeUp
       set    appAutoScroll,(iy+appFlags)
       ld     hl,appBackUpScreen
       call   DriverInit
       call   PumpOn
       ld     hl,8000h
       ld     b,82h
       call   SetupLog
       B_CALL GrBufClr
       ld     hl,plotSScreen
       ld     de,HubCB
       call   AutoSetup
       ret    c
       ld     a,cbNewData+1
       ld     (appBackUpScreen+128),a
       ld     b,1
       call   USBReqData
       ld     a,1
       ld     b,hubfeatPortPower
       call   SetHubPortFeature
       call   WaitTimer16
       ld     a,1
       ld     b,hubfeatPortReset
       call   SetHubPortFeature
HubLoop:
       ei
       halt
       ld     a,(appBackUpScreen+128)
       cp     cbNewData
       jr     nz,HubLoopSkip
       inc    a
       ld     (appBackUpScreen+128),a
       ld     bc,(appBackUpScreen+129)
       ld     hl,appBackUpScreen
HubDispLoop:
       ld     a,(hl)
       call   DispHexA
       inc    hl
       dec    bc
       ld     a,b
       or     c
       jr     nz,HubDispLoop
       ld     a,' '
       call   IPutC
HubLoopSkip:
       B_CALL GetCSC
       cp     skAdd
       jr     z,IncAddr
       cp     skSub
       jr     z,DecAddr
       cp     skEnter
       jr     z,TestDesc
       cp     skClear
       jr     nz,HubLoop
       call   KillUSBDevice
       call   KillDriver
       ret
IncAddr:
       in     a,(80h)
       inc    a
       out    (80h),a
       jr     HubLoop
DecAddr:
       in     a,(80h)
       dec    a
       out    (80h),a
       jr     HubLoop
TestDesc:
       xor    a
       out    (80h),a
       ld     b,descDevice
       ld     hl,plotSScreen
       push   hl
       call   ReadDescriptor
       pop    hl
       jr     c,HubLoop
       ld     b,16
$$:
       ld     a,(hl)
       call   DispHexA
       inc    hl
       djnz   $b
       ld     a,' '
       call   IPutC
       jr     HubLoop
       
HubCB:
       ld     a,b
       cp     cbNewData
       jr     nz,HubCBDone
       ld     (appBackUpScreen+128),a
       ld     (appBackUpScreen+129),de
HubCBDone:
       ld     b,1
       jp     USBReqData

 ENDIF

;Initialize
;Set Configuration
;Set Port Feature - Port Power
;Set Port Feature - Port Reset

